/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *                      Simon Heath - MSKCC                                 *
 *                                                                          *
 *                          August 2000                                     *
 *                                                                          *
 * loki_output.c:                                                           *
 *                                                                          *
 * Routines for sample output                                               *
 *                                                                          *
 ****************************************************************************/

#include <config.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

int n_cov_columns;

#include "utils.h"
#include "loki.h"
#include "loki_peel.h"
#include "version.h"
#include "loki_ibd.h"
#include "calc_var_locus.h"
#include "sample_rand.h"
#include "sample_nu.h"
#include "loki_output.h"

void Output_Sample_Aff(int lp,double *ss,double *ss2,int *perm,FILE *fptr)
{
	int k,k1,k2,k3;
	
	fprintf(fptr,"%d ",lp);
	for(k1=0;k1<n_links;k1++) {
		get_locuslist(perm,k1,&k2);
		if(analysis&NULL_ANALYSIS) {
			for(k3=0;k3<k2;k3++) if(perm[k3]==0) break;
			if(k3==k2) continue;
		}
		gnu_qsort(perm,(size_t)k2,(size_t)sizeof(int),cmp_loci_pos);
		for(k3=0;k3<k2;k3++) {
			k=perm[k3];
			if((analysis&NULL_ANALYSIS) && k) continue;
			fprintf(fptr,"%g ",ss[k]);
			if(ss2) fprintf(fptr,"%g ",ss2[k]);
		}
	}
	fputc('\n',fptr);
	fflush(fptr);
}

void Output_Sample_IBD(FILE *fptr,int lp)
{
	int i,i1,i2,j,k=0,l,link,comp;
	double *ibd1;
	
	fprintf(fptr,"Iteration %d\n",lp);
	for(link=0;link<n_links;link++) {
		if(linkage[link].ibd_list) {
			fprintf(fptr,"\n**Linkage group %s:\n",linkage[link].name);
			for(l=0;l<linkage[link].ibd_list->idx;l++) {
				fprintf(fptr,"\n**Position = %g\n",linkage[link].ibd_list->pos[l]);
				for(i1=comp=0;comp<n_comp;comp++) {
					ibd1=ibd[k][comp];
					for(i2=0,i=0;i<comp_size[comp];i++) {
							if(family_id) {
								print_orig_family(fptr,i1+i+1,0);
								fputc(' ',fptr);
							}
	  fprintf(fptr,"%d ",comp+1);
	  print_orig_id1(fptr,i1+i+1);
//	  for(j=0;j<3*jjbd[0][i1+i]+1;j++)fprintf(fptr," %g",(double)jjbd[j][i1+i]/(double)lp);
	  for(j=0;j<3*jjbd[0][i1+i][k]+1;j++)fprintf(fptr," %d",jjbd[j][i1+i][k]);
	  fputc('\n',fptr);
					}
					i1+=i;
				}
				k++;
			}
		}
	}
	k=0;
	for(link=0;link<n_links;link++) {
		if(linkage[link].ibd_list) {
			for(l=0;l<linkage[link].ibd_list->idx;l++) {
				for(i1=comp=0;comp<n_comp;comp++) {
					for(i2=0,i=0;i<comp_size[comp];i++) {
	  for(j=0;j<3*jjbd[0][i1+i][k]+1;j++)jjbd[j][i1+i][k]=0;
					}
					i1+=i;
				}
				k++;
			}
		}
	}
	fputc('\n',fptr);
	(void)fflush(fptr);
}

/* Version 2.1 output routine */
static void OutputSample_1(FILE *fptr,int lp)
{
	int i,j,k,l,type,nq,nq1,mn,mn1,grp;
	double z;
	
	nq=nq1=0;
	for(l=0;l<n_tloci;l++) if(tlocus[l].flag)	{
		nq++;
		if(tlocus[l].flag&TL_LINKED) nq1++;
	}
	fprintf(fptr,"%d %d %d %g %g %g",lp,nq,nq1,grand_mean,residual_var,tau);
	if(use_student_t) fprintf(fptr," %g",res_nu);
	if(polygenic_flag) fprintf(fptr," %g",additive_var);
	for(i=0;i<n_random;i++) fprintf(fptr," %g",c_var[i]);
	if(output_type==OUTPUT_TYPE_ORIGINAL) fprintf(fptr," %d %d",n_cov_columns,n_genetic_groups);
	for(i=0;i<model.n_terms;i++) if(model.term[i].out_flag) {
		type=model.term[i].vars[0].type;
		if(type&ST_ID) continue;
		k=model.term[i].vars[0].var_index;
		if(type&ST_MARKER) {
			for(grp=0;grp<n_genetic_groups;grp++)
				for(j=0;j<marker[k].n_alleles-1;j++) fprintf(fptr," %g",marker[k].freq[grp][j]);
		}
		if(!(type&ST_TRAITLOCUS))
		  for(j=0;j<model.term[i].df;j++) fprintf(fptr," %g",model.term[i].eff[j]);
		if(type&ST_MARKER) fprintf(fptr," %g",sqrt(calc_var_locus(k)));
	}
	for(l=0;l<n_tloci;l++) if(tlocus[l].flag) {
		if(tlocus[l].flag&TL_LINKED) {
			fprintf(fptr," %d",tlocus[l].link_group+1);
			for(k=0;k<=sex_map;k++) fprintf(fptr," %g",tlocus[l].pos[1-k]);
		} else {
			fputs(" 0",fptr);
			if(output_type==OUTPUT_TYPE_ORIGINAL) for(k=0;k<=sex_map;k++) fputs(" -1",fptr);
		}
		k=tlocus[l].n_alleles;
		for(grp=0;grp<n_genetic_groups;grp++)
		  for(j=0;j<k-1;j++) fprintf(fptr," %g",tlocus[l].freq[grp][j]);
		k=k*(k+1)/2-1;
		for(j=0;j<k;j++) fprintf(fptr," %g",tlocus[l].eff[j]);
		fprintf(fptr," %g",sqrt(tlocus[l].variance));
	}
	fputc('\n',fptr);
	if(debug_level&2) {
		mn=mn1=0;
		for(i=0;i<N_MOVE_STATS;i++) {
			z=move_stats[i].n?(double)move_stats[i].success/(double)move_stats[i].n:0.0;
			fputc(i?' ':'[',fptr);
			fprintf(fptr," %g",z);
			if(i<4) {
				mn+=move_stats[i].success;
				mn1+=move_stats[i].n;
			}
		}
		z=mn1?(double)mn/(double)mn1:0.0;
		fprintf(fptr," %g]\n",z);
	}
	(void)fflush(fptr);
}

/* Version 2.2/3 output routine */
static void OutputSample_2(FILE *fptr,int lp)
{
	int i,j,k,l,type,nq,nq1,mn,mn1,grp;
	double z,tot_gen_var;
	
	nq=nq1=0;
	tot_gen_var=0.0;
	for(l=0;l<n_tloci;l++) if(tlocus[l].flag) {
		nq++;
		if(tlocus[l].flag&TL_LINKED) nq1++;
		tot_gen_var+=tlocus[l].variance;
	}
	if(polygenic_flag) tot_gen_var+=additive_var;
	for(i=0;i<model.n_terms;i++) {
		k=model.term[i].vars[0].var_index;
		type=model.term[i].vars[0].type;
		if(type&ST_MARKER) tot_gen_var+=calc_var_locus(k);
	}
	switch(output_type) {
	 case 2:
		fprintf(fptr,"%d %g %g",lp,grand_mean,residual_var);
		if(use_student_t) fprintf(fptr," %g",res_nu);
		if(polygenic_flag) fprintf(fptr," %g",additive_var);
		for(i=0;i<n_random;i++) fprintf(fptr," %g",c_var[i]);
		break;
	 default:
		fprintf(fptr,"%d",lp);
		if(res_var_set!=1) fprintf(fptr," %g",residual_var);
		if(use_student_t) fprintf(fptr," %g",res_nu);
		fprintf(fptr," %g",tot_gen_var);
		if(polygenic_flag) fprintf(fptr," %g",additive_var);
		for(i=0;i<n_random;i++) fprintf(fptr," %g",c_var[i]);
		if(grand_mean_set!=1) fprintf(fptr," %g",grand_mean);
		break;
	}
	for(i=0;i<model.n_terms;i++) if(model.term[i].out_flag) {
		type=model.term[i].vars[0].type;
		if(type&ST_ID) continue;
		k=model.term[i].vars[0].var_index;
		if(type&ST_MARKER) {
			for(grp=0;grp<n_genetic_groups;grp++)
				for(j=0;j<marker[k].n_alleles-1;j++) fprintf(fptr," %g",marker[k].freq[grp][j]);
		}
		if(!(type&ST_TRAITLOCUS))
		  for(j=0;j<model.term[i].df;j++) fprintf(fptr," %g",model.term[i].eff[j]);
		if(type&ST_MARKER) {
			fprintf(fptr," %g",sqrt(calc_var_locus(k)));
		}
	}
	for(l=0;l<n_tloci;l++) if(tlocus[l].flag) {
		if(tlocus[l].flag&TL_LINKED) {
			fprintf(fptr," %d",tlocus[l].link_group+1);
			for(k=0;k<=sex_map;k++) fprintf(fptr," %g",tlocus[l].pos[1-k]);
		} else fputs(" 0",fptr);
		k=tlocus[l].n_alleles;
		for(grp=0;grp<n_genetic_groups;grp++)
		  for(j=0;j<k-1;j++) fprintf(fptr," %g",tlocus[l].freq[grp][j]);
		k=k*(k+1)/2-1;
		for(j=0;j<k;j++) fprintf(fptr," %g",tlocus[l].eff[j]);
		fprintf(fptr," %g",sqrt(tlocus[l].variance));
	}
	fputc('\n',fptr);
	if(debug_level&2) {
		mn=mn1=0;
		for(i=0;i<N_MOVE_STATS;i++) {
			z=move_stats[i].n?(double)move_stats[i].success/(double)move_stats[i].n:0.0;
			fputc(i?' ':'[',fptr);
			fprintf(fptr,"%g",z);
			if(i<4) {
				mn+=move_stats[i].success;
				mn1+=move_stats[i].n;
			}
		}
		z=mn1?(double)mn/(double)mn1:0.0;
		fprintf(fptr," %g]\n",z);
	}
	(void)fflush(fptr);
}

void OutputSample(FILE *fptr,int lp)
{
	switch(output_type) {
	 case 0:
	 case 1: 
		OutputSample_1(fptr,lp);
		break;
	 case 2:
	 case 3:
	 default:
		OutputSample_2(fptr,lp);
		break;
	}
}

void OutputFreqHeader(FILE *fptr)
{
	int i,j;
	
	fprintf(fptr,"Created by %s: %s",LOKI_NAME,ctime(&start_time));
	for(i=0;i<n_markers;i++) {
		fprintf(fptr,"%d %s",i,marker[i].name);
		if(marker[i].index) fprintf(fptr,"(%d)",marker[i].index);
		fputc(':',fptr);
		for(j=0;j<marker[i].n_alleles;j++) {
			fputc(' ',fptr);
			print_allele_type1(fptr,i,j);
		}
		fputc('\n',fptr);
	}
	fprintf(fptr,"No. genetic groups: %d\n",n_genetic_groups);
	fputs("--------------------\n",fptr);
	(void)fflush(fptr);
}

void OutputFreq(FILE *fptr,int lp)
{
	int i,j,k;
	
	for(k=0;k<n_genetic_groups;k++) {
		fprintf(fptr,"%d ",lp);
		for(i=0;i<n_markers;i++) for(j=0;j<marker[i].n_alleles;j++) fprintf(fptr,"%g ",marker[i].freq[k][j]);
		fputc('\n',fptr);
	}
	(void)fflush(fptr);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "OutputHeader"
void OutputHeader(FILE *fptr)
{
	int i,j,k,k1,k2,k3,type,n_all,grp,*perm=0;
	struct Variable *var,*group_var=0;
	
	for(i=0;i<n_id_records;i++) if(id_variable[i].type&ST_GROUP) {
		group_var=id_variable+i;
		break;
	}
	fprintf(fptr,"Created by %s: %s",LOKI_NAME,ctime(&start_time));
	if(!analysis) {
		fprintf(fptr,"Output format: %d\n",output_type);
		type=model.var.type;
		if(type) {
			fputs("Model: ",fptr);
			i=model.var.var_index;
			var=(type&ST_CONSTANT)?id_variable+i:nonid_variable+i;
			fputs(var->name,fptr);
			if(var->index) fprintf(fptr,"(%d)",var->index);
			fputs(" = ",fptr);
			for(i=0;i<model.n_terms;i++) {
				type=model.term[i].vars[0].type;
				if(i) fputs(" + ",fptr);
				if(type&ST_TRAITLOCUS) fputs("QTL",fptr);
				else if(type&ST_ID) fputs("ID\'",fptr);
				else {
					j=model.term[i].vars[0].var_index;
					if(type&ST_MARKER) {
						fputs(marker[j].name,fptr);
						if(marker[j].index) fprintf(fptr,"(%d)",marker[j].index);
					} else {
						var=(type&ST_CONSTANT)?id_variable+j:nonid_variable+j;
						fputs(var->name,fptr);
						if(var->index) fprintf(fptr,"(%d)",var->index);
						if(type&ST_RANDOM) fputc('\'',fptr);
					}
				}
			}
			fputc('\n',fptr);
		}
	}
	if(!(analysis&NULL_ANALYSIS)) {
		if(n_links) {
			if((n_tloci+n_markers) && !(perm=malloc(sizeof(int)*(n_tloci+n_markers)))) ABT_FUNC(MMsg);
			fputs("Linkage groups:\n",fptr);
			for(i=0;i<n_links;i++) {
				fprintf(fptr,"  %d: %s  Map range: ",i+1,linkage[i].name);
				for(k3=0;k3<=sex_map;k3++) fprintf(fptr," (%gcM to %gcM) ",linkage[i].r1[1-k3],linkage[i].r2[1-k3]);
				fputc('\n',fptr);
				if(perm) {
					get_locuslist(perm,i,&k2);
					gnu_qsort(perm,(size_t)k2,sizeof(int),cmp_loci_pos);
					for(k1=0;k1<k2;k1++) {
						k=perm[k1];
						if(k>=0) {
							if(marker[k].index) fprintf(fptr,"    %s(%d) - ",marker[k].name,marker[k].index);
							else fprintf(fptr,"    %s -",marker[k].name);
							for(k3=0;k3<=sex_map;k3++) fprintf(fptr," %g",marker[k].pos[1-k3]);
							fputc('\n',fptr);
						}
					}
				}
				if(linkage[i].ibd_list) {
					fprintf(fptr,"IBD Matrix estimated at position%s",linkage[i].ibd_list->idx==1?":":"s:");
					for(j=0;j<linkage[i].ibd_list->idx;j++) {
						fputc(j?',':' ',fptr);
						fprintf(fptr,"%g",linkage[i].ibd_list->pos[j]);
					}
					fputc('\n',fptr);
				}
			}
			if(perm) free(perm);
			fputs("Total Map Length:",fptr);
			for(k3=0;k3<=sex_map;k3++) fprintf(fptr," %gcM",total_maplength[1-k3]);
			fputc('\n',fptr);		}
	}
	if(!analysis) {
		for(k=i=0;i<model.n_terms;i++) if(model.term[i].out_flag) {
			type=model.term[i].vars[0].type;
			if(type&ST_ID) continue;
			if(!(type&ST_TRAITLOCUS)) k+=model.term[i].df;
			if(type&ST_MARKER) {
				k1=model.term[i].vars[0].var_index;
				k+=marker[k1].n_alleles;
			}
		}
		fputs("Output columns:\n",fptr);
		k1=0;
		fprintf(fptr,"  %d: Iteration count\n",++k1);
		if(output_type<=OUTPUT_VERSION_2_1) {
			fprintf(fptr,"  %d: No. QTL's in model\n",++k1);
			fprintf(fptr,"  %d: No. linked QTL's\n",++k1);
		}
		if(output_type==3) {
			if(res_var_set!=1) fprintf(fptr,"  %d: Residual variance\n",++k1);
			if(use_student_t) fprintf(fptr,"  %d: d.f. for student t distribution of residuals\n",++k1);
			fprintf(fptr,"  %d: Total genetic variance\n",++k1);
		} else {
			fprintf(fptr,"  %d: Grand mean\n",++k1);
			fprintf(fptr,"  %d: Residual variance\n",++k1);
			if(output_type<=OUTPUT_VERSION_2_1) fprintf(fptr,"  %d: Tau\n",++k1);
			if(use_student_t) fprintf(fptr,"  %d: d.f. for student t distribution of residuals\n",++k1);
		}
		if(polygenic_flag) fprintf(fptr,"  %d: Additive variance\n",++k1);
		for(j=0;j<model.n_terms;j++) {
			type=model.term[j].vars[0].type;
			if(type&ST_RANDOM) {
				k=model.term[j].vars[0].var_index;
				if(type&ST_CONSTANT) var=id_variable+k;
				else var=nonid_variable+k;
				fprintf(fptr,"  %d: Additional random variance for %s",++k1,var->name);
				if(var->index) fprintf(fptr,"(%d)",var->index);
				fputc('\n',fptr);
			}
		}
		if(output_type==OUTPUT_TYPE_ORIGINAL) {
			fprintf(fptr,"  %d: No. covariate columns\n",++k1);
			fprintf(fptr,"  %d: No. genetic groups\n",++k1);
		}
		k3=k1;
		if(k) {
			fputs(" covariate data:\n",fptr);
			if(output_type==3 && grand_mean_set!=1) fprintf(fptr,"  %d: Grand mean\n",++k1);
			for(i=0;i<model.n_terms;i++) if(model.term[i].out_flag)	{
				type=model.term[i].vars[0].type;
				if(type&ST_ID) continue;
				k=model.term[i].vars[0].var_index;
				if(type&ST_TRAITLOCUS) continue;
				if(type&ST_MARKER) {
					n_all=marker[k].n_alleles;
					for(grp=0;grp<n_genetic_groups;grp++) {
						for(j=0;j<n_all-1;j++) {
							fprintf(fptr,"  %d: %s",++k1,marker[k].name);
							if(marker[k].index) fprintf(fptr,"(%d)",marker[k].index);
							if(group_var) {
								fputs(" [Genetic group ",fptr);
								if(group_var->rec_flag==ST_STRING) fputs(group_var->recode[grp].string,fptr);
								else fprintf(fptr,"%d",group_var->recode[grp].value);
								fputc(']',fptr);
							}
							fputs(" freq. ",fptr);
							print_allele_type1(fptr,k,j);
							fputc('\n',fptr);
						}
					}
					for(j=1;j<n_all;j++) for(k2=0;k2<=j;k2++)	{
						fprintf(fptr,"  %d: %s",++k1,marker[k].name);
						if(marker[k].index) fprintf(fptr,"(%d)",marker[k].index);
						fputs(" effect ",fptr);
						print_allele_type1(fptr,k,k2);
						fputc(',',fptr);
						print_allele_type1(fptr,k,j);
						fputc('\n',fptr);
					}
					fprintf(fptr,"  %d: %s",++k1,marker[k].name);
					if(marker[k].index) fprintf(fptr,"(%d)",marker[k].index);
					fputs(" size\n",fptr);
				} else {	
					if(type&ST_CONSTANT) var=id_variable+k;
					else var=nonid_variable+k;
					for(j=0;j<model.term[i].df;j++) {
						fprintf(fptr,"  %d: %s",++k1,var->name);
						if(var->index) fprintf(fptr,"(%d)",var->index);
						fputs(" effect ",fptr);
						if(var->type&ST_FACTOR)	{
							if(var->rec_flag==ST_STRING) fputs(var->recode[j].string,fptr);
							else fprintf(fptr,"%d",var->recode[j].value);
						}
						fputc('\n',fptr);
					}
				}
			}
		}
		if(tlocus && max_tloci) {
			fputs(" QTL data blocks:\n  linkage group\n",fptr);
			if(output_type==OUTPUT_TYPE_ORIGINAL) {
 				if(sex_map) fputs("  male position\n  female position\n",fptr);
				else fputs("  position\n",fptr);
			} else {
 				if(sex_map) fputs("  [male position if linked]\n  [female position if linked]\n",fptr);
				else fputs("  [position if linked]\n",fptr);
			}
			for(grp=0;grp<n_genetic_groups;grp++) {
				if(group_var) {
					fputs("  [Genetic group ",fptr);
					if(group_var->rec_flag==ST_STRING) fputs(group_var->recode[grp].string,fptr);
					else fprintf(fptr,"%d",group_var->recode[grp].value);
					fputc(']',fptr);
				}
				fputs(" freq. 2\n",fptr);
			}
			fputs("  effect 1,2\n  effect 2,2\n  size\n",fptr);
			if(min_tloci==max_tloci) fprintf(fptr,"No. QTL: %d\n",max_tloci);
			else {
				fprintf(fptr,"Number of QTL: %d to %d\n",min_tloci,max_tloci);
				if(tloci_mean_set) fprintf(fptr,"Mean of poisson prior on QTL number: %g\n",tloci_mean);
			}
		}
		if(res_var_set==1) fprintf(fptr,"Residual variance: %g\n",residual_var);
		if(grand_mean_set==1) fprintf(fptr,"Grand mean: %g\n",grand_mean);
		fprintf(fptr,"Tau Mode: %d\nTau Beta: %g\n",tau_mode,tau_beta);
		fprintf(fptr,"No. fixed output columns: %d\n",k1);
		if(no_overdominant) fputs("Over-dominant QTLs not allowed\n",fptr);
		k1-=k3;
		n_cov_columns=k1;
	} else if(!(analysis&ESTIMATE_IBD)) {
		fputs("Affected only ",fptr);
		if(analysis&IBD_ANALYSIS) fputs("IBD Analysis\n",fptr);
		else if(analysis&IBS_ANALYSIS) fputs("IBS	Analysis\n",fptr);
		else fputc('\n',fptr);
	}
	if(lm_ratio>0.0) fprintf(fptr,"LM ratio: %g\n",lm_ratio);
	if(l_prob>0.0) fprintf(fptr,"l_prob: %g, l_cycles: %d\n",l_prob,l_cycles);
	fprintf(fptr,"No. genetic groups: %d\n",n_genetic_groups);
	if(sex_map) fputs("Sex specific map\n",fptr);
	fputs("--------------------\n",fptr);
	(void)fflush(fptr);
}

