#! /usr/bin/perl 
#-w
#
# Loki_QTLc-1.0: perl script to help building the file containing
# MCMC-based QTL-coefficients
# Yurii Aulchenko
# Institute of Cytology and Genetics SD RAS, Novosibirsk, Russia
# Created: 26/01/2001
# Last updated: 26/01/2001
#

#
# Read pedigree data in Haley's format
#
print "input pedigree file name: ";$pedfile=<STDIN>;chomp $pedfile;
open PEDF,"<$pedfile" or die "Cannt open file \"$pedfile\": $!\n";
print "input marker genotypes file name: ";$mrkfile=<STDIN>;chomp $mrkfile;
open MRKF,"<$mrkfile" or die "Cannt open file \"$mrkfile\": $!\n";
print "input map information file name: ";$mapfile=<STDIN>;chomp $mapfile;
open MAPF,"<$mapfile" or die "Cannt open file \"$mapfile\": $!\n";
print "input total number of MCMC iterates: ";$niter=<STDIN>;chomp $niter; 
print "input number of MCMC relaxation iterates: ";$s_iter=<STDIN>;chomp $s_iter; 
print "input step (cM) used for calculation of QTL-coefficients: ";$step1=<STDIN>;chomp $step1; 
die "Starting iteration >= total iterantions\n" if ($niter <= $s_iter);
die "Step <= 0\n" if ($step1 <= 0);
$npr=0;
while (<PEDF>) {
 $temp = 1../^$/;
 if (!$temp) {$npr--;last;};
 $peddat[$npr] = $_;
 $npr++;
} 
close PEDF;
@iff2 = 1..$npr;
@ifp = (0) x $npr;
for ($i=0;$i<$npr;$i++) {
 @tline=split(/\s+/,$peddat[$i]);
 $tline[0] ? $iff2[$i]=$tline[0] : $iff2[$i]=$tline[1];
}

#
# Check for extra-informtaion
#
# Mark extra fields indicating number of F-2 offspring
print "Re-writing pedigree in Loki format...";
@index = 1..$npr;
for ($i=0;$i<=$npr;$i++) {
 @nel=split(/\s+/,$peddat[$i]);
 $nel=@nel;
 $index[$i] = 0 if ($nel-2 <= 0);
 if(($nel - 2 > 0) and ($nel[2] != 0)) {
  @bug = splice @nel, 5;
  push @nel, (3,@bug);
  $ndat=join " ",@nel;
  if(@nel<=6){
   $peddat[$i]=$ndat." \*\n";
   $iff2[$i]=0;
  } else{
   $peddat[$i]=$ndat."\n";}
 } else {
  @tline=split(/\s+/,$peddat[$i]);
  splice @tline,0,1 if !$tline[0];
  $ifp[$i]=$tline[4];
  $nel[2]="\*";$nel[3]="\*";
  $ndat=join " ",@nel;
  $peddat[$i]=$ndat." \*\n";
  $iff2[$i]=0;
 }
# print "$i; $index[$i]; $iff2[$i]; $ifp[$i] $peddat[$i]";
}
# Mark extra fields containing parents, described more then once
for ($i=0;$i<$npr;$i++) {
 @nel2 = split /\s+/,$peddat[$i];
 if ($index[$i]) {
  $id = $nel2[1];
  for ($j=$i+1;$j<=$npr;$j++) {
   if ($index[$j]) {
    @nel3 = split /\s+/,$peddat[$j];
    $id1 = $nel3[1];
    $index[$j] = 0 if ($id1 == $id);
   }
  }
 }
}
#
# Write out the data
#
$nids=0;
open OUTF,">lokiped.lqt" or die "Cannt open file \"lokiped.lqt\": $!\n";
for ($i=0;$i<$npr;$i++) {
 if ($index[$i]){print OUTF "$peddat[$i]";$nids++;} 
}
close(OUTF);
print "done\n";
#
# Read MAP information
#
print "Re-writing map data in Loki format...";
$mline1=<MAPF>;
$mline2=<MAPF>;
$mline3=<MAPF>;
close MAPF;
#print "$mline1";
#print "$mline2";
#print "$mline3";
@mline1=split /[^0-9\.]+/,$mline1;
#for($i=0;$i<@mline1;$i++) {splice(@mline1,$i,1) if (@mline1[$i]=="");}
splice(@mline1,0,1) if (!$mline1[0]);
@mline2=split /[^0-9\.]+/,$mline2;
#for($i=0;$i<@mline2;$i++) {splice(@mline2,$i,1) if (@mline2[$i]=="");}
splice(@mline2,0,1) if (!$mline2[0]);
@mline3=split /[^0-9\.]+/,$mline3;
#for($i=0;$i<@mline3;$i++) {splice(@mline3,$i,1) if (@mline3[$i]=="");}
splice(@mline3,0,1) if (!$mline3[0]);
if ($mline1[0]!=1) 
 {die "number of chromosomes under analysis is restricted to 1\n";}
$nmark=$mline2[0];
$nm_file=$mline1[1];
#print "n_and_n: $nmark $nm_file";print ": $_\n" foreach(@mline3);
if ($nmark>$nm_file) 
 {die "Declared number of markers conflict (MAP file, line 1 and 2)";}
if ($nmark!=(@mline2-1)) 
 {die "declared and real number of markers conflict (MAP file, line 2)\n";}
if ($nmark!=(@mline3-1)) 
 {die "declared and real number of markers conflict (MAP file, line 3)\n";}
#print "\n";
@m_numbers = splice @mline2, 1;
#foreach(@temp1) {print"\"$_\" ";}
@m_pos = splice @mline3, 1;
#foreach(@temp2) {print"\"$_\" ";}
for($i=1;$i<@m_pos;$i++){$m_pos[$i]+=$m_pos[$i-1];}

#
# Make control file
#

open OUTF,">control.lqt" or die "Cannt open file \"control.lqt\": $!\n";
select(OUTF);
$~=CONTROL1;
write; 
print OUTF "Array hap1_left($nm_file),hap1_right($nm_file),D1G($nm_file)\n";
$~=CONTROL2;
write; 
print OUTF "file DataDir+\"lokimrk.lqt\",id,(hap1_left(i),hap1_right(i),i=1,$nm_file)\n";
$~=CONTROL3;
write; 
print OUTF "while(i<=$nm_file)\n";
$~=CONTROL4;
write; 
print OUTF "Link \"Chromosome 1\"";
for ($i=0;$i<$nmark;$i++) {
 print OUTF ",D1G($m_numbers[$i])";}
print OUTF "\n";
close OUTF;

#
# Make parameter file
#
open OUTF,">param.lqt" or die "Cannt open file \"control.lqt\": $!\n";
print OUTF "Iterations $niter\n";
print OUTF "Start Output $s_iter,100\n";
$fr=$niter-$s_iter;
print OUTF "Output frequency $fr,100\n";
print OUTF "Output sx,parity\n";
print OUTF "estimate ibd \"Chromosome 1\"";
$i=$m_pos[0];
while ($i<=$m_pos[$nmark-1]){
print OUTF ",$i";
$i+=$step1;
}
print OUTF "\n";
print OUTF "analyze ibd\n";
print OUTF "Seedfile \"seedfile\",1\n";
for ($i=0;$i<$nmark;$i++) {
 print OUTF "Position D1G($m_numbers[$i])  $m_pos[$i]\n";}
close(OUTF);
select(STDOUT);
print "done\n";
#
# Rewrite marker file
#
print "Re-writing marker data in Loki format...";
open OUTF,">lokimrk.lqt" or die "Cannt open file \"lokimrk.lqt\": $!\n";
while (<MRKF>) {
 $temp = $_;
 $temp =~ s/\s0/ \*/g;
 print OUTF "$temp";
}
close OUTF;
close MRKF;
print "done\n";

#
# Start ./prep and ./loki
#

print "Stariting ./prep...\n";
#$args=("./prep control.lqt 1>stdout.lqt 2>stderr.lqt");
$args=("./prep control.lqt");
#system($args) == 0 or die "System $args failed: $?;\nsee stderr.lqt and stdout.lqt for details\n";
system($args) == 0 or die "System $args failed: $?\n";
print "done\n";
print "Starting ./loki-qtlc...\n";
$args=("./loki-qtlc param.lqt");
#system($args) == 0 or die "System $args failed: $?;\nsee stderr.lqt and stdout.lqt for details\n";
system($args) == 0 or die "System $args failed: $?\n";
print "... done\n";

#
# Split loki.ibd at "position" points
#
print "Splitting loki.ibd at position points...\n ";
open LI,"loki.ibd" or die "Cannt open \"loki.ibd\": $!";
$redir=0;
$allpos="";
while (<LI>) {
 $temp=$_;
 $temp2=$_;
 if ($temp =~ m/osition/){
  @temp1=split /\s+/,$_;
  $pos=$temp1[2];
  print "$pos\n";
  $allpos.="$pos ";
  $redir=1;
  open CCF,">count.${pos}.lqt" or die "Cannt open temporary count data file";
 } elsif ($temp2 =~ m/^S/) {
   close CCF if $redir==1;
   $redir=0;
 } else {
   print CCF "$_" if $redir;
 }
}
close CCF if $redir;
close LI;
print "...done\n";
@allpos=split/\s+/,$allpos;

#
# Make control file for coe_lo programm
#
$np1=0;$np2=0;
for($i=0;$i<$npr;$i++){
 $np1++ if($index[$i]!=0 and $ifp[$i]==1);
}
for($i=0;$i<$npr;$i++){
 $np2++ if($index[$i]!=0 and $ifp[$i]==2);
}
open CCL,">calcbl.lqt" or die "Cannot open \"calbl.dat\": $!";
print CCL <<header;
#                                                      
# line 1: number of individuals
# line 2: number of founders of breed 1, founders' ids
# line 3: number of founders of breed 2, founders' ids
#
header
print CCL "$nids\n";
print CCL "$np1 ";
for($i=0;$i<$npr;$i++){
 if($index[$i]!=0 and $ifp[$i]==1){
  @tline=split(/\s+/,$peddat[$i]);
  splice @tline,0,1 if !$tline[0];
  $iddd=$tline[0];
  print CCL "$iddd "; 
 }
}
print CCL "\n";
print CCL "$np2 ";
for($i=0;$i<$npr;$i++){
 if($index[$i]!=0 and $ifp[$i]==2){
  @tline=split(/\s+/,$peddat[$i]);
  splice @tline,0,1 if !$tline[0];
  $iddd=$tline[0];
  print CCL "$iddd "; 
 }
}
close CCL;

#
# Processing splitted count data
#
print "processing splitted count data files and\n \tcalculating temporary coeff-files...\n";
for ($i=0;$i<@allpos;$i++) {
$pos=$allpos[$i];
print " $pos\n";
$args=("./coe_lo calcbl.lqt count.${pos}.lqt coeff.${pos}.lqt");
system($args) == 0 or die "System $args failed: $?\n";
}
print "...done\n";

#
# Compose QTL-c file
#

print "starting composition of single QTL-coefficients file...\n";
for ($i=0;$i<@iff2;$i++) {
 if (!$iff2[$i]){splice @iff2,$i,1;$i--} 
}
$nf2=@iff2;
$ncp=@allpos;
$ncp1=$ncp-1;
open LC,">loki.qtlc" or die "Cannt open \"loki.qtlc\": $!";
print LC " $nf2 1\n";print LC " 1 $ncp1\n";
print "processing ID...\n ";
for ($j=0;$j<$nf2;$j++) {
 $cid=$iff2[$j];print "$cid\n";
 for ($i=0;$i<@allpos;$i++) {
  $pos=$allpos[$i];
  open TF,"<coeff.${pos}.lqt";
  @tfile=<TF>;
  @line=grep(/ $cid /,@tfile);$line=$line[0];
  close TF;
  @lel=split /\s+/,$line;
  splice @lel,0,1 if (!$lel[0]);
  print LC " $lel[2] $lel[3]\n";
 }
}
close LC;
open LCF,">loki.qtlc.full" or die "Cannt open \"loki.qtlc.full\": $!";
 for ($i=0;$i<@allpos;$i++) {
  print LCF "Position = $allpos[$i]\n";
  open TF,"<coeff.${pos}.lqt";
  @tfile=<TF>;
  print LCF "@tfile\n";
  close TF;
 }
close LCF;
print "...done\n";

print `rm *.lqt`;
print `rm log`;
print `rm *~`;
print `rm loki.???`;

# control file format
format CONTROL1=
/*****************************************************************************
 *                                                                           *
 *                    Control file for the Loki-QTLc   			     *
 *                                                                           *
 *****************************************************************************/
.
format CONTROL2=
Filter "/usr/bin/gzip"    # Compress binary output files (recommended)
DataDir = "" # Where the data files are located
Missing "*"       # Define "0" as a missing value for all id variables
file DataDir+"lokiped.lqt",id,fa,mo,sx,gengr,qt
GROUP gengr
.
format CONTROL3=
/* Link haplotypes to the appropriate markers */
i=1
do
 Marker Locus D1G(i)[hap1_left(i),hap1_right(i)]
 i=i+1
.
format CONTROL4=
Trait Locus QTL     # Define QTL as 1 or more trait loci
pedigree id,fa,mo         # id(1), id(2), and id(3) defined as id, sire, dam
log "log"           # Name for log file
sex sx 1,2          # sx determines sex: 1=Male, 2=Female
model qt=QTL
.
###################################
